# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple
from pywwise.decorators import callback
from pywwise.enums import ELogChannel, ELogSeverity
from pywwise.statics import EnumStatics
from pywwise.structs import LogItem


class Log:
    """ak.wwise.core.log"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.item_added = _RefEvent(ELogChannel, LogItem)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_log_itemadded.html
        \nSent when an item is added to the log. To retrieve the complete log, refer to `ak.wwise.core.log.get`.
        \n**Event Data**:
        \n- The channel on which the item was added.
        \n- The item added to the log.
        """
        
        self._item_added = self._client.subscribe("ak.wwise.core.log.itemAdded", self._on_item_added)
    
    @callback
    def _on_item_added(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `itemAdded` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        channel = EnumStatics.from_value(ELogChannel, kwargs["channel"])
        item = LogItem(severity=EnumStatics.from_value(ELogSeverity, kwargs["item"]["severity"]),
                       time=kwargs["item"]["time"],
                       id=kwargs["item"]["messageId"],
                       description=kwargs["item"]["message"])
        event(channel, item)
    
    def add_item(self, message: str, severity: ELogSeverity = ELogSeverity.MESSAGE,
                 channel: ELogChannel = ELogChannel.GENERAL) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_log_additem.html \n
        Adds a new item to the logs on the specified channel.
        :param message: The text of the message to add.
        :param severity: The severity of the message to add.
        :param channel: The channel on which to add the message.
        :return: Whether the call succeeded.
        """
        args = {"message": message, "severity": severity.value, "channel": channel.value}
        return self._client.call("ak.wwise.core.log.addItem", args) is not None
    
    def clear(self, channels: ELogChannel | ListOrTuple[ELogChannel] = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_log_clear.html \n
        Clears the logs on one or more channels.
        :param channels: The channel or channels to clear. Duplicates will be ignored. If `None`, all channels will be
                         cleared.
        :return: Whether the call succeeded.
        """
        if channels is None:
            channels = tuple(EnumStatics.get_all_members(ELogChannel))
        elif isinstance(channels, ELogChannel):
            channels = (channels,)  # convert single value to a collection, so the loop below still works
        
        returns = list[bool]()
        for channel in channels:
            args = {"channel": channel}
            returns.append(self._client.call("ak.wwise.core.log.clear", args) is not None)
        
        return all(returns)
    
    def get(self, channel: ELogChannel = ELogChannel.GENERAL) -> tuple[LogItem, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_log_get.html \n
        Retrieves the latest log for a specific channel. Refer to `ak.wwise.core.log.item_added` to be
        notified when an item is added to the log. The log is empty when used in WwiseConsole.
        :param channel: The channel to clear.
        :return: A tuple of log items. May be empty.
        """
        args = {"channel": channel.value}
        result = self._client.call("ak.wwise.core.log.get", args)
        return tuple(LogItem(severity=EnumStatics.from_value(ELogSeverity, item["severity"]),
                             time=item["time"],
                             id=item["messageId"],
                             description=item["message"])
                     for item in result.get("items", ()))



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import (ELogicalOperator, ESampleRateConversionQuality, ESearchCriteriaContainerType,
                           ESearchCriteriaCurveType, ESearchCriteriaCurveUsage, ESearchCriteriaLfeOption,
                           ESearchCriteriaMode, ESearchCriteriaNumericOperator, ESearchCriteriaOtherChannelsCountOption,
                           ESearchCriteriaRtpcOperator, ESearchCriteriaSoundbankReferences, ESearchCriteriaSoundType,
                           ESearchCriteriaSampleRateConversionType, ESearchCriteriaStateProperty,
                           ESearchCriteriaSwitchingOperator, ESearchCriteriaUsingOperator)
from pywwise.objects.abc import WwiseObject


class SearchCriteria(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_searchcriteria.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SEARCH_CRITERIA`.
    """
    attenuation = WwiseProperty[str]("Attenuation", str)
    conditional_operator = WwiseProperty[ELogicalOperator]("ConditionalOperator", ELogicalOperator)
    conversion = WwiseProperty[str]("Conversion", str)
    curve_type = WwiseProperty[ESearchCriteriaCurveType]("CurveType", ESearchCriteriaCurveType)
    curve_usage = WwiseProperty[ESearchCriteriaCurveUsage]("CurveUsage", ESearchCriteriaCurveUsage)
    default_switch_state = WwiseProperty[str]("DefaultSwitchState", str)
    effect = WwiseProperty[str]("Effect", str)
    effect_class_id = WwiseProperty[str]("EffectClassID", str)
    events_referenced_in_operator = WwiseProperty[ESearchCriteriaSoundbankReferences](
        "EventReferencedInOperator", ESearchCriteriaSoundbankReferences)
    game_parameter = WwiseProperty[str]("GameParameter", str)
    has_empty_assignation = WwiseProperty[bool]("HasEmptyAssignation", bool)
    inclusion = WwiseProperty[bool]("Inclusion", bool)
    is_checked = WwiseProperty[bool]("IsChecked", bool)
    is_linked = WwiseProperty[bool]("IsLinked", bool)
    is_source_of_override = WwiseProperty[bool]("IsSourceOfOverride", bool)
    is_special_search = WwiseProperty[bool]("IsSpecialSearch", bool)
    is_used = WwiseProperty[bool]("IsUsed", bool)
    left_value = WwiseProperty[float]("LeftValue", float)
    logical_operator = WwiseProperty[ELogicalOperator]("LogicalOperator", ELogicalOperator)
    metadata = WwiseProperty[str]("Metadata", str)
    metadata_class_id = WwiseProperty[str]("MetadataClassID", str)
    mode = WwiseProperty[ESearchCriteriaMode]("Mode", ESearchCriteriaMode)
    music_segment = WwiseProperty[str]("MusicSegment", str)
    numeric_operator = WwiseProperty[ESearchCriteriaNumericOperator]("NumericOperator", ESearchCriteriaNumericOperator)
    object_referenced = WwiseProperty[str]("ObjectReferenced", str)
    output_bus = WwiseProperty[str]("OutputBus", str)
    property_name = WwiseProperty[str]("PropertyName", str)
    rtpc_operator = WwiseProperty[ESearchCriteriaRtpcOperator]("RTPCOperator", ESearchCriteriaRtpcOperator)
    right_value = WwiseProperty[float]("RightValue", float)
    rnd_seq_container_type = WwiseProperty[ESearchCriteriaContainerType](
        "RndSeqContainerType", ESearchCriteriaContainerType)
    sample_rate_conversion_quality = WwiseProperty[ESampleRateConversionQuality](
        "SRConversionQuality", ESampleRateConversionQuality)
    sample_rate_conversion_type = WwiseProperty[ESearchCriteriaSampleRateConversionType](
        "SampleRateConversionType", ESearchCriteriaSampleRateConversionType)
    sound_type = WwiseProperty[ESearchCriteriaSoundType]("SoundType", ESearchCriteriaSoundType)
    source_audio_format = WwiseProperty[str]("SourceAudioFormat", str)
    source_channels_lfe_option = WwiseProperty[ESearchCriteriaLfeOption](
        "SourceChannelsLFEOption", ESearchCriteriaLfeOption)
    source_channels_others_count = WwiseProperty[int]("SourceChannelsOthersCount", int)
    source_channels_others_option = WwiseProperty[ESearchCriteriaOtherChannelsCountOption](
        "SourceChannelsOthersOption", ESearchCriteriaOtherChannelsCountOption)
    source_language = WwiseProperty[str]("SourceLanguage", str)
    source_type = WwiseProperty[str]("SourceType", str)
    state = WwiseProperty[str]("State", str)
    state_group = WwiseProperty[str]("StateGroup", str)
    state_property_usage = WwiseProperty[ESearchCriteriaStateProperty](
        "StatePropertyUsage", ESearchCriteriaStateProperty)
    switch_group = WwiseProperty[str]("SwitchGroup", str)
    switching_operator = WwiseProperty[ESearchCriteriaSwitchingOperator](
        "SwitchingOperator", ESearchCriteriaSwitchingOperator)
    trigger = WwiseProperty[str]("Trigger", str)
    u_int_value = WwiseProperty[int]("UIntValue", int)
    user_string = WwiseProperty[str]("UserString", str)
    using_operator = WwiseProperty[ESearchCriteriaUsingOperator]("UsingOperator", ESearchCriteriaUsingOperator)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path as _Path
from re import Pattern as _Pattern
from typing import (List as _List, Set as _Set, Tuple as _Tuple, TypeAlias as _TypeAlias, TypeVar as _TypeVar,
                    Union as _Union)

# region BuiltIn

SystemPath: _TypeAlias = _Path
"""Represents a filesystem path."""

RegexPattern: _TypeAlias = _Pattern
"""Represents a Regex pattern."""

UnionType: _TypeVar = _TypeVar("UnionType")
"""Type variable for Unions."""

ListOrTuple: _TypeAlias = _Union[_List[UnionType], _Tuple[UnionType, ...]]
"""Custom type representing both `List` and `Tuple`."""

ListOrTupleOrSet: _TypeAlias = _Union[_List[UnionType], _Tuple[UnionType, ...], _Set[UnionType]]
"""Custom type representing `List`, `Tuple`, and `Set`."""

# endregion



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import EventHandler as _EventHandler, WaapiClient as _WaapiClient

from pywwise.decorators import callback
from pywwise.enums import EObjectType, ETransportExecuteActions, ETransportState
from pywwise.primitives import GameObjectID, GUID, Name, ProjectPath
from pywwise.statics import EnumStatics
from pywwise.structs import TransportObjectInfo


class Transport:
    """ak.wwise.core.transport"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.state_changed = _RefEvent(int, ETransportState, GUID, GameObjectID)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_statechanged.html
        \nSent when the transport's state has changed. A transport needs to be created first, using `create`.
        \n**Event Data**:
        \n- The ID (unsigned integer 32-bit) of the watched transport.
        \n- A string containing the old name.
        \n- The GUID of the object controlled by the transport object.
        \n- The GameObjectID used by the transport object. An invalid (`-1`) ID just means no game object was used.
        """
        
        self._state_changed = dict[int, _EventHandler]()  # keys are transport IDs; subscriptions are dynamic
    
    @callback
    def _on_state_changed(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `stateChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        kwargs["state"] = EnumStatics.from_value(ETransportState, kwargs["state"])
        kwargs["gameObject"] = GameObjectID(kwargs.get("gameObject", -1))  # -1 means the ID is invalid
        event(kwargs["transport"], kwargs["state"], kwargs["object"], kwargs["gameObject"])
    
    def create(self, wwise_obj: GUID | tuple[EObjectType, Name] | ProjectPath,
               game_obj: GameObjectID = None) -> int:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_create.html \n
        Creates a transport object for the given Wwise object. The return transport object can be used to
        play, stop, pause and resume the Wwise object via the other transport functions.
        :param wwise_obj: The ID (GUID), typed name, or project path of the object to control via the transport object.
        :param game_obj: The game object to use for playback.
        :return: Transport object ID to be used with all other ak.wwise.core.transport functions. Returns -1 if function
                 call failed.
        """
        is_typed_name = isinstance(wwise_obj, tuple)
        args = {"object": wwise_obj if not is_typed_name else f"{wwise_obj[0].get_type_name()}:{wwise_obj[1]}"}
        if game_obj is not None:
            args["gameObject"] = game_obj
        
        results = self._client.call("ak.wwise.core.transport.create", args)
        if results is None or results.get("transport") is None:
            return -1
        
        transport_id = results["transport"]
        if self._state_changed.get(transport_id) is None:  # subscription doesn't exist, so we need to set one up
            self._state_changed[transport_id] = self._client.subscribe("ak.wwise.core.transport.stateChanged",
                                                                       {"transport": transport_id})
        
        return -1 if results is None else results.get("transport", -1)
    
    def destroy(self, transport_id: int) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_destroy.html \n
        Destroys the given transport object.
        :param transport_id: The transport object ID to be used with all other ak.wwise.core.transport functions. Make
                             sure you create a transport object before calling this function.
        :return: True if the transport object was destroyed, False otherwise.
        """
        args = {"transport": transport_id}
        
        results = self._client.call("ak.wwise.core.transport.destroy", args)
        if results is None:
            return False
        
        event_handler = self._state_changed.get(transport_id)
        if event_handler is not None:  # unsubscription should only happen if a subscription exists
            self._client.unsubscribe(event_handler)
        
        return True
    
    def execute_action(self, transport_action: ETransportExecuteActions, transport_id: int = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_executeaction.html \n
        Executes an action on the given transport object, or all transport objects if none is specified.
        :param transport_action: The action to execute. Use ETransportExecuteActions to see the available actions.
        :param transport_id: The transport object ID to be used with all other ak.wwise.core.transport functions. Make
                             sure you create a transport object before calling this function.
        :return: True if the action was executed, False otherwise.
        """
        args = {"action": transport_action, **({"transport": transport_id} if transport_id is not None else dict())}
        return self._client.call("ak.wwise.core.transport.execute", args) is not None
    
    def get_list(self) -> tuple[TransportObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_getlist.html \n
        Returns the list of transport objects.
        :return: A list of transport objects.
        """
        results = self._client.call("ak.wwise.core.transport.getList", {})
        
        results = results.get("list") if results is not None else None
        
        if results is None:
            return ()
        
        transport_objects = list[TransportObjectInfo]()
        
        for result in results:
            obj = GUID(result["object"])
            game_object = int(result["gameObject"])
            transport = int(result["transport"])
            transport_objects.append(TransportObjectInfo(obj, game_object, transport))
        
        return tuple(transport_objects)
    
    def get_state(self, transport_id: int) -> ETransportState:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_getstate.html \n
        Gets the state of the given transport object.
        :param transport_id: The transport object ID to be used with all other ak.wwise.core.transport functions. Make
                             sure you create a transport object before calling this function.
        :return: An enum value that represents the state of the transport object. Return value can either be: Playing,
                 Stopped, Paused, or None. In case its none, the method call failed.
        """
        args = {"transport": transport_id}
        
        result = self._client.call("ak.wwise.core.transport.getState", args)
        
        if result is None:
            return ETransportState.NONE
        
        # from_value will [attempt to] convert any value to the specified Enum subclass.
        return EnumStatics.from_value(ETransportState, result.get("state", "None"))
    
    def prepare(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_prepare.html \n
        Prepare the object and its dependencies for playback. Use this function before calling
        `PostEventSync` or `PostMIDIOnEventSync` from `IAkGlobalPluginContext`.
        :param obj: The GUID, typed name, or project path of the object to prepare. Only object types that have
                    globally-unique names are supported (e.g. EObjectType.EVENT).
        :return: True if the call was successful, False otherwise.
        """
        is_name = isinstance(obj, tuple)
        args = {"object": obj if not is_name else f"{obj[0].get_type_name()}:{obj[1]}"}
        return self._client.call("ak.wwise.core.transport.prepare", args) is not None
    
    def use_originals(self, flag: bool = False) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_transport_useoriginals.html \n
        Sets the Original/Converted transport toggle globally. This allows playing the original or the
        converted sound files.
        :param flag: True to enable original files, False to use converted files. If no value is provided, system will
                     default to using converted files.
        :return: True if the call was successful, False otherwise.
        """
        args = {"enable": flag}
        return self._client.call("ak.wwise.core.transport.useOriginals", args) is not None



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient


class Waapi:
    """ak.wwise.ui.waapi"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
    
    def get_functions(self) -> tuple[str, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_waapi_getfunctions.html \n
        Retrieves the list of functions.
        :return: A tuple containing the URI of all WAAPI functions.
        """
        results = self._client.call("ak.wwise.waapi.getFunctions")
        return tuple(results.get("functions")) if results is not None else ()
    
    def get_schema(self, uri: str) -> dict | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_waapi_getschema.html \n
        Retrieves the JSON schema of a Waapi URI.
        :param uri: The URI to get the schema for (e.g. "ak.wwise.waapi.getSchema" gets the schema for this function).
        :return: If the call succeeded, a JSON-like dictionary representing the schema of the specified URI; else, None.
        """
        return self._client.call("ak.wwise.waapi.getSchema", {"uri": uri})
    
    def get_topics(self) -> tuple[str, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_waapi_gettopics.html \n
        Retrieves the list of topics to which a client can subscribe.
        :return: A tuple containing the URI of all WAAPI topics.
        """
        results = self._client.call("ak.wwise.waapi.getTopics")
        return tuple(results.get("topics")) if results is not None else ()



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.enums import EObjectType
from pywwise.primitives import GUID, Name, ProjectPath


class AudioSourcePeaks:
    """ak.wwise.core.audioSourcePeaks"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
    
    def get_min_max_peaks_in_region(self, source: GUID | Name | ProjectPath, time_from: float, time_to: float,
                                    num_peaks: int, cross_channel_peaks: bool = False) -> tuple[
        tuple[bytes, ...], int, float, float, float, str]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audiosourcepeaks_getminmaxpeaksinregion.html \n
        Gets the min/max peak pairs, in the given region of an audio source, as a collection of binary strings (one per
        channel). The strings are base-64 encoded, 16-bit signed int arrays, with min and max values being interleaved.
        If getCrossChannelPeaks is true, only one binary string represents the peaks across all channels globally.
        :param source: The GUID, name, or project path of the Audio Source to get the min and max peaks from.
        :param time_from: The start time, in seconds, of the section of the audio source for which peaks are required.
                          This number must be smaller than timeTo. Range: [0,*]
        :param time_to: The end time, in seconds, of the section of the audio source for which peaks are required. This
                        number must be larger than timeFrom. Range: [0,*]
        :param num_peaks: The number of peaks that are required (minimum 1). Range: [1,4294967295]
        :param cross_channel_peaks: When true, peaks are calculated globally across channels, instead of per channel.
        :returns: A tuple containing the binary strings representing the peaks, the amount of channels, the maximum
                  absolute value, the length of the tuple, the size of the data in the tuple, and the description of
                  the channel configuration. **Refer to the Audiokinetic documentation for more detailed information**.
                  NOTE: if the call fails, this function will return an empty tuple.
        """
        time_from = max(time_from, 0)
        time_to = max(time_to, 0)
        num_peaks = max(num_peaks, 1)
        
        if time_from >= time_to:
            raise ValueError("The value of `time_from` must be smaller than the value of `time_to`.")
        
        args = {"object": f"{EObjectType.AUDIO_SOURCE}:{source}" if isinstance(source, Name) else source,
                "timeFrom": time_from, "timeTo": time_to, "numPeaks": num_peaks,
                "getCrossChannelPeaks": cross_channel_peaks}
        
        result = self._client.call("ak.wwise.core.audioSourcePeaks.getMinMaxPeaksInRegion", args)
        
        if result is None:
            return tuple[tuple[bytes, ...], int, float, float, float, str]()  # empty
        return (result["peaksBinaryStrings"], int(result["numChannels"]), result["maxAbsValue"],
                result["peaksArrayLength"], result["peaksDataSize"], result["channelConfig"])
    
    def get_min_max_peaks_in_trimmed_region(self, source: GUID | Name | ProjectPath, num_peaks: int,
                                            cross_channel_peaks: bool = False) -> tuple[
        tuple[bytes, ...], int, float, float, float, str]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audiosourcepeaks_getminmaxpeaksintrimmedregion.html \n
        Gets the min/max peak pairs in the entire trimmed region of an audio source, for each channel, as an array of
        binary strings (one per channel). The strings are base-64 encoded, 16-bit signed int arrays, with min and max
        values being interleaved. If getCrossChannelPeaks is true, there is only one binary string representing peaks
        across all channels globally.
        :param source: The GUID, name, or project path of the Audio Source to get the min and max peaks from.
        :param num_peaks: The number of peaks that are required (minimum 1). Range: [1,4294967295]
        :param cross_channel_peaks: When true, peaks are calculated globally across channels, instead of per channel.
        :returns: A tuple containing the binary strings representing the peaks, the amount of channels, the maximum
                  absolute value, the length of the tuple, the size of the data in the tuple, and the description of
                  the channel configuration. **Refer to the Audiokinetic documentation for more detailed information**.
                  NOTE: if the call fails, this function will return an empty tuple.
        """
        num_peaks = max(num_peaks, 1)
        
        args = {"object": f"{EObjectType.AUDIO_SOURCE}:{source}" if isinstance(source, Name) else source,
                "numPeaks": num_peaks, "getCrossChannelPeaks": cross_channel_peaks}
        
        result = self._client.call("ak.wwise.core.audioSourcePeaks.getMinMaxPeaksInTrimmedRegion", args)
        
        if result is None:
            return tuple[tuple[bytes, ...], int, float, float, float, str]()  # empty
        return (result["peaksBinaryStrings"], int(result["numChannels"]), result["maxAbsValue"],
                result["peaksArrayLength"], result["peaksDataSize"], result["channelConfig"])



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.objects.abc import WwiseObject


class ExternalSource(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_externalsource.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.EXTERNAL_SOURCE`.
    """
    pass  # There are no properties, currently.



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.waapi.ak.wwise.ui.commands import Commands
from pywwise.waapi.ak.wwise.ui.project import Project
from pywwise.waapi.ak.wwise.ui.ui import UI



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EMeterMode, EMeterScope
from pywwise.objects.abc import WwiseObject


class WwiseMeter(WwiseObject):  # TODO: add RTPC property once available
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_meter.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_METER`.
    """
    apply_downstream_volume = WwiseProperty[bool]("ApplyDownstreamVolume", bool)
    attack_time = WwiseProperty[float]("AttackTime", float)
    hold = WwiseProperty[float]("Hold", float)
    infinite_hold = WwiseProperty[bool]("InfiniteHold", bool)
    max = WwiseProperty[float]("Max", float)
    meter_mode = WwiseProperty[EMeterMode]("MeterMode", EMeterMode)
    meter_scope = WwiseProperty[EMeterScope]("MeterScope", EMeterScope)
    min = WwiseProperty[float]("Min", float)
    release_time = WwiseProperty[float]("ReleaseTime", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import SystemPath
from pywwise.structs import PlatformInfo


class Project:
    """ak.wwise.ui.project"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
    
    def close(self, bypass_save: bool = True) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_project_close.html \n
        Closes the current project.
        :param bypass_save: Indicates if the user should not be prompted to save the current project.
                            **Defaults to true.**
        :return: True if there was a project open, false otherwise. Note that if there was no project open,
                 no `ak.wwise.core.project.pre_closed` or `ak.wwise.core.project.post_closed` event is issued.
        """
        args = {"bypassSave": bypass_save}
        result = self._client.call("ak.wwise.ui.project.close", args)
        return result.get("hadProjectOpen")
    
    def create(self, path: SystemPath, platforms: set[PlatformInfo] = None, languages: set[str] = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_project_create.html \n
        Creates, saves and opens new empty project, specified by path and platform. The project has no
        factory setting WorkUnit. Please refer to `ak.wwise.core.project.loaded` for further explanations
        on how to be notified when the operation has completed.
        :param path: The path to the project WPROJ file. The path must use the same name for the WPROJ and the parent
                     directory folder. For example: `C:/PyWwise/Projects/MYPROJECT/MYPROJECT.wproj`.
        :param platforms: Specifies the platform or platforms supported by the new project. If not specified, only
                          Windows is used. Duplicates are not allowed; platforms should have **unique names**.
        :param languages: Array of languages to creates for this project. If not specified, the English(US) language is
                          created. When multiple languages are specified, the first one becomes the default language.
        :return: Whether the project creation succeeded. This is done by checking if the specified path now
                 exists.
        """
        args = {"path": str(path)}
        
        if platforms is not None:
            args["platforms"] = list()
            for platform in platforms:
                args["platforms"].append({"name": platform.name, "basePlatform": platform.base})
        if languages is not None:
            args["languages"] = list(languages)
        
        self._client.call("ak.wwise.ui.project.create", args)
        return path.exists()
    
    def open(self, path: SystemPath, is_migration_required: bool = False, bypass_save: bool = True,
             version_control_auto_checkout: bool = True) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_project_open.html \n
        Opens a project, specified by path. Please refer to `ak.wwise.core.project.loaded` for further
        explanations on how to be notified when the operation has completed.
        :param path: The path to the project WPROJ file. For using WAAPI on Mac, please refer to Using WAAPI on Mac.
        :param is_migration_required: Whether migration is required or not.
        :param bypass_save: Indicates if the user should not be prompted to save the current project.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a Checkout source control
                                              operation for affected work units and for the project. Only supported
                                              in Wwise 2023 or above.
        :return: Returns whether the project was open.
        """
        if not path.exists():
            return False
        
        migration_action = "migrate" if is_migration_required else "fail"
        
        args = {"path": str(path),
                "onMigrationRequired": migration_action,
                "bypassSave": bypass_save,
                **({"autoCheckoutToSourceControl": False} if not version_control_auto_checkout else {})}
        
        return self._client.call("ak.wwise.ui.project.open", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pywwise.objects.types.rtpc import Rtpc

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour, EEnvelopeTriggerOn, EModulatorScopeLimited
from pywwise.objects.abc import WwiseObject


class ModulatorTime(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_modulatortime.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MODULATOR_TIME`.
    """
    colour = WwiseProperty[EColour]("Color", EColour)
    envelope_stop_playback = WwiseProperty[bool]("EnvelopeStopPlayback", bool)
    envelope_trigger_on = WwiseProperty[EEnvelopeTriggerOn]("EnvelopeTriggerOn", EEnvelopeTriggerOn)
    modulator_scope = WwiseProperty[EModulatorScopeLimited]("ModulatorScope", EModulatorScopeLimited)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    rtpc: WwiseProperty[tuple[Rtpc, ...]] = WwiseProperty("RTPC", tuple)  # Using `:` to avoid circular imports.
    time_mod_duration = WwiseProperty[float]("TimeModDuration", float)
    time_mod_initial_delay = WwiseProperty[float]("TimeModInitialDelay", float)
    time_mod_loops = WwiseProperty[int]("TimeModLoops", int)
    time_mod_playback_rate = WwiseProperty[float]("TimeModPlaybackRate", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EBuiltInGameParameter, EColour, EInterpolationMode
from pywwise.objects.abc import WwiseObject


class GameParameter(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_gameparameter.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.GAME_PARAMETER`.
    """
    bind_to_built_in_param = WwiseProperty[EBuiltInGameParameter]("BindToBuiltInParam", EBuiltInGameParameter)
    colour = WwiseProperty[EColour]("Color", EColour)
    filter_time_down = WwiseProperty[float]("FilterTimeDown", float)
    filter_time_up = WwiseProperty[float]("FilterTimeUp", float)
    initial_value = WwiseProperty[float]("InitialValue", float)
    max = WwiseProperty[float]("Max", float)
    min = WwiseProperty[float]("Min", float)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    rtpc_ramping = WwiseProperty[EInterpolationMode]("RTPCRamping", EInterpolationMode)
    simulation_value = WwiseProperty[float]("SimulationValue", float)
    slew_rate_down = WwiseProperty[float]("SlewRateDown", float)
    slew_rate_up = WwiseProperty[float]("SlewRateUp", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple, SystemPath
from pywwise.decorators import callback
from pywwise.enums import EAudioImportOperation, EImportOperation, ELogSeverity, EObjectType, EReturnOptions
from pywwise.primitives import GUID, Name, ProjectPath
from pywwise.statics import EnumStatics
from pywwise.structs import AudioImportEntry, ConversionLogItem, WwiseObjectInfo


class Audio:
    """ak.wwise.core.audio"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.imported = _RefEvent(EImportOperation, tuple[WwiseObjectInfo, ...], tuple[SystemPath, ...])
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_imported.html
        \nSent at the end of an import operation.
        \n**Event Data**:
        \n- The operation applied for the import.
        \n- A tuple of WwiseObjectInfo instances, representing objects created as part of the import operation.
        \n- A tuple of SystemPath instances, representing the paths of the imported assets.
        """
        
        imported_args = {"return": [EReturnOptions.GUID.value, EReturnOptions.NAME.value,
                                    EReturnOptions.TYPE.value, EReturnOptions.PATH.value]}
        self._imported = self._client.subscribe("ak.wwise.core.audio.imported", self._on_imported,
                                                imported_args)
    
    @callback
    def _on_imported(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `imported` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        objects = list[WwiseObjectInfo]()
        for obj in kwargs.get("objects", dict()):
            objects.append(WwiseObjectInfo.from_dict(obj))
        event(EnumStatics.from_value(EImportOperation, kwargs["operation"]), tuple(objects),
              tuple([SystemPath(file) for file in kwargs.get("files", ())]))
    
    def convert(self, objects: ListOrTuple[GUID | Name | ProjectPath], platforms: ListOrTuple[GUID | Name],
                languages: ListOrTuple[Name]) -> tuple[ConversionLogItem, ...]:
        """
        https://www.audiokinetic.com/en/library/2024.1.0_8598/?source=SDK&id=ak_wwise_core_audio_convert.html \n
        Creates converted audio files. When errors occur, this function returns a list of messages with corresponding
        levels of severity. The converted audio files are located within the ".cache" folder in the Wwise project
        root folder.
        :param objects: An array of object GUIDs, unique Names, or Project Paths.
        :param platforms: An array of platform GUIDs or unique Names.
        :param languages: An array of language unique Names.
        :return: A tuple of logged entries with associated messages and severities. If empty, the conversion(s) worked
                 without any errors, warnings, etc.
        """
        args = {"objects": objects, "platforms": platforms, "languages": languages}
        result: dict[str, list[dict[str, str]]] = self._client.call("ak.wwise.core.audio.convert", args)
        return tuple(ConversionLogItem(EnumStatics.from_value(ELogSeverity, error["severity"]),
                                       error.get("message", "")) for error in result.get("errors", ()))
    
    # "import" is a reserved keyword, so function name does not match that of WAAPI
    def import_files(self, imports: ListOrTuple[AudioImportEntry],
                     operation: EAudioImportOperation = EAudioImportOperation.USE_EXISTING,
                     version_control_auto_add: bool = True,
                     version_control_auto_checkout: bool = True,
                     platform: Name | GUID = None,
                     language: Name | GUID = None) -> tuple[WwiseObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_import.html \n
        Creates Wwise objects and imports audio files. This function does not return an error when something fails
        during the import process, please refer to the log for the result of each import command. This function uses
        the same import processor available through the Tab Delimited import in the Audio File Importer. The function
        returns an array of all objects created, replaced or re-used. Use the options to specify how the objects are
        returned.
        :param imports: An audio import entry. The members of this object are combined with those of "default", with
                        this object's members having precedence. If a property specified in defaults is also specified
                        here, the value in this object will take precedence over the one in defaults.
        :param operation: Determines how import object creation is performed. Make use of EImportOperation to select
                          the desired import operation.
        :param version_control_auto_add: Determines if Wwise automatically performs a source control Add operation on
                                         the imported files. Defaults to true. Only supported in Wwise 2023 or above.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a source control Checkout
                                              operation (when applicable) on the modified files. Defaults to true. Only
                                              supported in Wwise 2023 or above.
        :param platform: Determines what platform the Wwise object is returned. This is an optional argument. When not
                         specified, the current platform is used.
        :param language: Determines the language to be used.
        :return: A tuple of WwiseObjectInfo instances, representing the objects that were created and/or edited.
        """
        args = {"importOperation": operation,
                "imports": [entry.dictionary for entry in imports],
                **({"autoAddToSourceControl": False} if not version_control_auto_add else {}),
                **({"autoCheckOutToSourceControl": False} if not version_control_auto_checkout else {})}
        
        options = {"return": EReturnOptions.get_defaults(),
                   **({"platform": platform} if platform is not None else {}),
                   **({"language": language} if language is not None else {})}
        
        result = self._client.call("ak.wwise.core.audio.import", args, options=options)
        objects = result.get("objects", ()) if isinstance(result, dict) else ()
        
        return tuple(WwiseObjectInfo(GUID(obj["id"]),
                                     Name(obj["name"]),
                                     EObjectType.from_type_name(obj["type"]),
                                     ProjectPath(obj["path"]))
                     for obj in objects)
    
    def import_tab_delimited(self,
                             tsv_file: SystemPath,
                             language: Name | GUID,
                             platform: Name | GUID = None,
                             operation: EImportOperation = EAudioImportOperation.USE_EXISTING,
                             root_path: GUID | tuple[EObjectType, Name] | ProjectPath = None,
                             version_control_auto_add: bool = True,
                             version_control_auto_checkout: bool = True) -> tuple[WwiseObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_importtabdelimited.html \n
        Scripted object creation and audio file import from a tab-delimited file.
        :param tsv_file: Location of tab-delimited import file.
        :param operation: Determines how import object creation is performed. Make use of EImportOperation to select
                          your desired import operation.
        :param language: Imports language for audio file import (taken from the project's defined languages,
                         found in the WPROJ file LanguageList).
        :param platform: The platform to use during the import operations. If not specified, the currently-active
                         platform will be used.
        :param root_path: The GUID, typed name, or path used as root relative object paths. Although using a typed name
                          is supported, it will only work with globally-unique names (e.g. Event names).
        :param version_control_auto_add: Determines if Wwise automatically performs a source control Add operation on
                                         the imported files. Defaults to true. Only supported in Wwise 2023 or above.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a source control Checkout
                                              operation (when applicable) on the modified files. Defaults to true. Only
                                              supported in Wwise 2023 or above.
        """
        args = {"importLanguage": language,
                "importOperation": operation,
                "importFile": str(tsv_file),
                "importLocation": root_path,
                **({"autoAddToSourceControl": False} if not version_control_auto_add else {}),
                **({"autoCheckOutToSourceControl": False} if not version_control_auto_checkout else {})}
        
        options = {"return": EReturnOptions.get_defaults(),
                   **({"platform": platform} if platform is not None else {}),
                   **({"language": language} if language is not None else {})}
        
        result = self._client.call("ak.wwise.core.audio.importTabDelimited", args, options=options)
        objects = result.get("objects", ()) if isinstance(result, dict) else ()
        
        return tuple(WwiseObjectInfo(GUID(obj["id"]),
                                     Name(obj["name"]),
                                     EObjectType.from_type_name(obj["type"]),
                                     ProjectPath(obj["path"]))
                     for obj in objects)
    
    def mute(self, objs: ListOrTuple[GUID | tuple[EObjectType, Name] | ProjectPath], value: bool) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_mute.html \n
        Mute or unmute objects.
        :param objs: The GUID, Name, or Project Path of the objects to mute. Using Names is supported for object
                     types that have unique names (e.g. Buses, Events, etc.), but it is still NOT recommended. If
                     you choose to use Names, you will have to specify the type of the object
                     (e.g. `Name("Event:Play_Footstep")`).
        :param value: Whether the objects should be muted or not.
        :return: Whether the call succeeded. True does not necessarily mean objects were muted/unmuted successfully.
        """
        objs = [obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}" for obj in objs]
        args = {"objects": objs, "value": value}
        return self._client.call("ak.wwise.core.audio.mute", args) is not None
    
    def reset_mute(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_resetmute.html \n
        Unmute all muted objects.
        :return: Whether the call succeeded. True does not necessarily mean objects were unmuted successfully.
        """
        return self._client.call("ak.wwise.core.audio.resetMute") is not None
    
    def reset_solo(self) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_resetsolo.html \n
        Unsolo all soloed objects.
        :return: Whether the call succeeded. True does not necessarily mean objects were unsoloed successfully.
        """
        return self._client.call("ak.wwise.core.audio.resetSolo") is not None
    
    def set_conversion_plugin(self, conversion: GUID | Name | ProjectPath, plugin: Name, platform: GUID | Name) -> bool:
        """
        https://www.audiokinetic.com/en/library/2024.1.0_8598/?source=SDK&id=ak_wwise_core_audio_setconversionplugin.html \n
        :param conversion: The GUID, Name, or Project Path or a Conversion shareset.
        :param plugin: The name of the plugin to use for future conversions (e.g. Vorbis).
        :param platform: The GUID or Name of the platform to which the settings apply.
        :return: Whether the call succeeded.
        """
        conversion = conversion if not isinstance(conversion, Name) else f"{EObjectType.CONVERSION}:{conversion}"
        args = {"conversion": conversion, "plugin": plugin, "platform": platform}
        return self._client.call("ak.wwise.core.audio.setConversionPlugin", args) is not None
    
    def solo(self, objs: ListOrTuple[GUID | tuple[EObjectType, Name] | ProjectPath], value: bool) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_audio_solo.html \n
        Solos an object.
        :param objs: The GUIDs, typed names, or project paths of the objects to solo. Using names is supported only for
                     object types that have unique names (e.g. Buses, Events, etc.), but it is still NOT recommended.
        :param value: Whether the object is soloed or not. 1 is true, 0 is false.
        :return: Whether the call succeeded. True does not necessarily mean objects were soloed/unsoloed successfully.
        """
        objs = [obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}" for obj in objs]
        args = {"objects": objs, "value": value}
        return self._client.call("ak.wwise.core.audio.solo", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import (E3DPosition, E3DSpatialization, EColour, EDiscardBehaviour, ELoudnessNormalizationType,
                           EMidiPlayOnNoteType, EOverLimitBehaviour, EScope, EVirtualVoiceBehaviour,
                           EVirtualVoiceQueueBehaviour)
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.attenuation import Attenuation
from pywwise.objects.types.aux_bus import AuxBus
from pywwise.objects.types.bus import Bus
from pywwise.objects.types.conversion import Conversion
from pywwise.objects.types.effect_slot import EffectSlot
from pywwise.objects.types.rtpc import Rtpc
from pywwise.primitives import GUID


class ActorMixer(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_actormixer.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.ACTOR_MIXER`.
    """
    position_3d = WwiseProperty[E3DPosition]("3DPosition", E3DPosition)
    spatialization_3d = WwiseProperty[E3DSpatialization]("3DSpatialization", E3DSpatialization)
    attenuation = WwiseProperty[Attenuation]("Attenuation", Attenuation)
    attenuation_distance_scaling = WwiseProperty[float]("AttenuationDistanceScaling", float)
    below_threshold_behaviour = WwiseProperty[EVirtualVoiceBehaviour]("BelowThresholdBehavior", EVirtualVoiceBehaviour)
    bypass_effect = WwiseProperty[bool]("BypassEffect", bool)
    centre_percentage = WwiseProperty[float]("CenterPercentage", float)
    colour = WwiseProperty[EColour]("Color", EColour)
    conversion = WwiseProperty[Conversion]("Conversion", Conversion)
    effects = WwiseProperty[tuple[EffectSlot, ...]]("Effects", tuple)
    enable_attenuation = WwiseProperty[bool]("EnableAttenuation", bool)
    enable_diffraction = WwiseProperty[bool]("EnableDiffraction", bool)
    enable_loudness_normalization = WwiseProperty[bool]("EnableLoudnessNormalization", bool)
    enable_midi_note_tracking = WwiseProperty[bool]("EnableMidiNoteTracking", bool)
    game_aux_send_hpf = WwiseProperty[int]("GameAuxSendHPF", int)
    game_aux_send_lpf = WwiseProperty[int]("GameAuxSendLPF", int)
    game_aux_send_volume = WwiseProperty[float]("GameAuxSendVolume", float)
    hdr_active_range = WwiseProperty[float]("HdrActiveRange", float)
    hdr_enable_envelope = WwiseProperty[bool]("HdrEnableEnvelope", bool)
    hdr_envelope_sensitivity = WwiseProperty[float]("HdrEnvelopeSensitivity", float)
    highpass = WwiseProperty[int]("Highpass", int)
    hold_emitter_position_orientation = WwiseProperty[bool]("HoldEmitterPositionOrientation", bool)
    hold_listener_orientation = WwiseProperty[bool]("HoldListenerOrientation", bool)
    ignore_parent_max_sound_instance = WwiseProperty[bool]("IgnoreParentMaxSoundInstance", bool)
    inclusion = WwiseProperty[bool]("Inclusion", bool)
    is_global_limit = WwiseProperty[EScope]("IsGlobalLimit", EScope)
    listener_relative_routing = WwiseProperty[bool]("ListenerRelativeRouting", bool)
    loudness_normalization_target = WwiseProperty[float]("LoudnessNormalizationTarget", float)
    loudness_normalization_type = WwiseProperty[ELoudnessNormalizationType](
        "LoudnessNormalizationType", ELoudnessNormalizationType)
    lowpass = WwiseProperty[int]("Lowpass", int)
    make_up_gain = WwiseProperty[float]("MakeUpGain", float)
    max_reached_behaviour = WwiseProperty[EDiscardBehaviour]("MaxReachedBehavior", EDiscardBehaviour)
    max_sound_per_instance = WwiseProperty[int]("MaxSoundPerInstance", int)
    metadata = WwiseProperty[GUID]("Metadata", GUID)
    midi_break_on_note_off = WwiseProperty[bool]("MidiBreakOnNoteOff", bool)
    midi_channel_filter = WwiseProperty[int]("MidiChannelFilter", int)
    midi_key_filter_max = WwiseProperty[int]("MidiKeyFilterMax", int)
    midi_key_filter_min = WwiseProperty[int]("MidiKeyFilterMin", int)
    midi_play_on_note_type = WwiseProperty[EMidiPlayOnNoteType]("MidiPlayOnNoteType", EMidiPlayOnNoteType)
    midi_tracking_root_note = WwiseProperty[int]("MidiTrackingRootNote", int)
    midi_transposition = WwiseProperty[int]("MidiTransposition", int)
    midi_velocity_filter_max = WwiseProperty[int]("MidiVelocityFilterMax", int)
    midi_velocity_filter_min = WwiseProperty[int]("MidiVelocityFilterMin", int)
    midi_velocity_offset = WwiseProperty[int]("MidiVelocityOffset", int)
    output_bus = WwiseProperty[Bus]("OutputBus", Bus)
    output_bus_highpass = WwiseProperty[int]("OutputBusHighpass", int)
    output_bus_lowpass = WwiseProperty[int]("OutputBusLowpass", int)
    output_bus_volume = WwiseProperty[float]("OutputBusVolume", float)
    over_limit_behaviour = WwiseProperty[EOverLimitBehaviour]("OverLimitBehavior", EOverLimitBehaviour)
    override_analysis = WwiseProperty[bool]("OverrideAnalysis", bool)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    override_conversion = WwiseProperty[bool]("OverrideConversion", bool)
    override_early_reflections = WwiseProperty[bool]("OverrideEarlyReflections", bool)
    override_effect = WwiseProperty[bool]("OverrideEffect", bool)
    override_game_aux_sends = WwiseProperty[bool]("OverrideGameAuxSends", bool)
    override_hdr_envelope = WwiseProperty[bool]("OverrideHdrEnvelope", bool)
    override_metadata = WwiseProperty[bool]("OverrideMetadata", bool)
    override_midi_events_behaviour = WwiseProperty[bool]("OverrideMidiEventsBehavior", bool)
    override_midi_note_tracking = WwiseProperty[bool]("OverrideMidiNoteTracking", bool)
    override_output = WwiseProperty[bool]("OverrideOutput", bool)
    override_positioning = WwiseProperty[bool]("OverridePositioning", bool)
    override_priority = WwiseProperty[bool]("OverridePriority", bool)
    override_user_aux_sends = WwiseProperty[bool]("OverrideUserAuxSends", bool)
    override_virtual_voice = WwiseProperty[bool]("OverrideVirtualVoice", bool)
    pitch = WwiseProperty[int]("Pitch", int)
    priority = WwiseProperty[int]("Priority", int)
    priority_distance_factor = WwiseProperty[bool]("PriorityDistanceFactor", bool)
    priority_distance_offset = WwiseProperty[int]("PriorityDistanceOffset", int)
    rtpc = WwiseProperty[tuple[Rtpc, ...]]("RTPC", tuple)
    reflections_aux_send = WwiseProperty[AuxBus]("ReflectionsAuxSend", AuxBus)
    reflections_volume = WwiseProperty[float]("ReflectionsVolume", float)
    speaker_panning = WwiseProperty[int]("SpeakerPanning", int)
    speaker_panning_3d_spatialization_mix = WwiseProperty[int]("SpeakerPanning3DSpatializationMix", int)
    use_game_aux_sends = WwiseProperty[bool]("UseGameAuxSends", bool)
    use_max_sound_per_instance = WwiseProperty[bool]("UseMaxSoundPerInstance", bool)
    user_aux_send_0 = WwiseProperty[AuxBus]("UserAuxSend0", AuxBus)
    user_aux_send_1 = WwiseProperty[AuxBus]("UserAuxSend1", AuxBus)
    user_aux_send_2 = WwiseProperty[AuxBus]("UserAuxSend2", AuxBus)
    user_aux_send_3 = WwiseProperty[AuxBus]("UserAuxSend3", AuxBus)
    user_aux_send_hpf_0 = WwiseProperty[int]("UserAuxSendHPF0", int)
    user_aux_send_hpf_1 = WwiseProperty[int]("UserAuxSendHPF1", int)
    user_aux_send_hpf_2 = WwiseProperty[int]("UserAuxSendHPF2", int)
    user_aux_send_hpf_3 = WwiseProperty[int]("UserAuxSendHPF3", int)
    user_aux_send_lpf_0 = WwiseProperty[int]("UserAuxSendLPF0", int)
    user_aux_send_lpf_1 = WwiseProperty[int]("UserAuxSendLPF1", int)
    user_aux_send_lpf_2 = WwiseProperty[int]("UserAuxSendLPF2", int)
    user_aux_send_lpf_3 = WwiseProperty[int]("UserAuxSendLPF3", int)
    user_aux_send_volume_0 = WwiseProperty[float]("UserAuxSendVolume0", float)
    user_aux_send_volume_1 = WwiseProperty[float]("UserAuxSendVolume1", float)
    user_aux_send_volume_2 = WwiseProperty[float]("UserAuxSendVolume2", float)
    user_aux_send_volume_3 = WwiseProperty[float]("UserAuxSendVolume3", float)
    virtual_voice_queue_behaviour = WwiseProperty[EVirtualVoiceQueueBehaviour](
        "VirtualVoiceQueueBehavior", EVirtualVoiceQueueBehaviour)
    volume = WwiseProperty[float]("Volume", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Any as _Any, TYPE_CHECKING as _TYPE_CHECKING, TypeVar as _TypeVar

if _TYPE_CHECKING:
    from pywwise.structs import WwiseObjectInfo

from abc import ABC as _ABC
from enum import Enum as _Enum
from types import NoneType as _NoneType

from pywwise.enums import EObjectType
from pywwise.primitives import GUID, Name, ProjectPath
from pywwise.waapi.ak.ak import WwiseConnection


class WwiseObject(_ABC):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wobjects_index.html \n
    The base class for any class that serves as interface for getting/setting properties on Wwise objects.
    """
    
    def __init__(self, guid: GUID | WwiseObjectInfo, ak: WwiseConnection,
                 platform: GUID | Name | _NoneType = None):
        """
        Uses a GUID to initialize a strongly-typed dynamic object, capable of fetching information from Wwise as needed.
        :param guid: If you may also pass a `WwiseObjectInfo` instance - this function will extract only the GUID.
        :param platform: If you want your object to only be used on a specific platform, specify which one here.
        """
        self._ak: WwiseConnection = ak
        self._guid: GUID = guid if isinstance(guid, GUID) else getattr(guid, "guid", GUID.get_null())
        self._query: str = f"$ from object \"{self._guid}\" take 1"
        self._platform = platform
    
    def get_property(self, name: str, default: _Any = None) -> _NoneType | bool | int | float | str | GUID | _Enum:
        """
        Gets the value of a property, reference, or list from this object in Wwise.
        :param name: The name of the property, reference, or list.
        :param default: The default value, in case retrieving the value fails.
        :return: The value of the property. This *can* be `None`.
        """
        info = self._ak.wwise.core.object.get(self._query, (name,))[0]
        return info.other.get(name, default)
    
    def set_property(self, name: str, value: _Any, is_reference: bool = False):
        if not is_reference:
            self._ak.wwise.core.object.set_property(self._guid, name, value, self._platform)
        else:
            self._ak.wwise.core.object.set_reference(self._guid, name, value, self._platform)
    
    @property
    def guid(self) -> GUID:
        """
        Get GUID.
        :return: Current GUID.
        """
        return self._guid
    
    @property
    def name(self) -> Name:
        """
        Get name.
        :return: Current name.
        """
        return self._ak.wwise.core.object.get(self._query)[0].name
    
    @name.setter
    def name(self, name: Name | str):
        """
        Set name.
        :param name: New name.
        """
        self._ak.wwise.core.object.set_name(self._guid, name)
    
    @property
    def path(self) -> ProjectPath:
        """
        Get path.
        :return: Current path.
        """
        return self._ak.wwise.core.object.get(self._query)[0].path
    
    @path.setter
    def path(self, path: ProjectPath):
        """
        Set path.
        :param path: New path.
        """
        if path[-1] == '\\':  # ProjectPath always uses `\\` instead of `/`.
            path = path[:-1]  # Remove slash.
        tokens = path.split('\\')
        if tokens[-1] == self.name:  # We only need the path up to the parent.
            path = tokens[:-1]  # Remove name.
        self._ak.wwise.core.object.move(self.guid, path)
    
    @property
    def type(self) -> EObjectType:
        """
        Get type.
        :return: The type.
        """
        return self._ak.wwise.core.object.get(self._query)[0].type


WwiseObjectType = _TypeVar("WwiseObjectType", bound=WwiseObject)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.waapi.ak.wwise.console.project import Project as _Project


class Console:
    """ak.wwise.console"""
    
    def __init__(self, client: _WaapiClient, is_console_instance: bool = False):
        """
        Constructor.
        :param client: The WAAPI client to use.
        :param is_console_instance: Should be set to true if the instance of Wwise is running in a console window.
        """
        self._client = client
        self.project = _Project(self._client, is_console_instance)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple, SystemPath
from pywwise.decorators import callback
from pywwise.enums import (EGeneratedSoundBankType, EInclusionFilter, EInclusionOperation, ELogSeverity, EObjectType,
                           EReturnOptions)
from pywwise.primitives import GUID, Name, ProjectPath, ShortID
from pywwise.statics import EnumStatics
from pywwise.structs import (ExternalSourceInfo, LogItem, PluginLibraryInfo, SoundBankData, SoundBankGenerationInfo,
                             SoundBankInclusion, SoundBankInfo)


class SoundBank:
    """ak.wwise.core.soundbank"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.generated = _RefEvent(SoundBankGenerationInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_generated.html
        \nSent when a SoundBank is generated. Can multi-trigger during SoundBank generation, per bank and per platform.
        \n**Event Data**:
        \n- A SoundBankGenerationInfo instance, which contains information about the generated SoundBank.
        """
        
        generated_args = {"infoFile": True, "bankData": True, "pluginInfo": True,
                          "return": [EReturnOptions.GUID, EReturnOptions.NAME,
                                     EReturnOptions.TYPE, EReturnOptions.PATH]}
        
        self._generated = self._client.subscribe("ak.wwise.core.soundbank.generated",
                                                 self._on_generated, generated_args)
        
        self.generation_done = _RefEvent(tuple[LogItem])
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_generationdone.html
        \nSent when all SoundBanks are generated. Do not use to check if `ak.wwise.core.soundbank.generate` has completed.
        \n**Event Data**:
        \n- A tuple of LogItems representing the SoundBank generation log. Empty when used in WwiseConsole.
        """
        
        self._generation_done = self._client.subscribe("ak.wwise.core.soundbank.generationDone",
                                                       self._on_generation_done)
    
    @callback
    def _on_generated(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `generated` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        # Key: soundBank
        bank = kwargs.get("soundbank", dict())
        guid = GUID(bank.get("id", GUID.get_null()))
        name = Name(bank.get("name", Name.get_null()))
        path = ProjectPath(bank.get("path", ProjectPath.get_null()))
        
        # Key: bankInfo
        info = kwargs.get("bankInfo", list[dict]())
        info = info[0] if info else dict()
        short = ShortID(info.get("Id", ShortID.get_null()))
        file = str(info.get("Path", ""))
        btype = EnumStatics.from_value(EGeneratedSoundBankType, info.get("Type", ""))
        bhash = GUID(info.get("Hash", GUID.get_null()))
        
        # Key: bankInfo.Media
        media = info.get("Media", dict())
        
        # Key: bankInfo.ExternalSources
        externals = info.get("ExternalSources", dict())
        
        # Key: bankInfo.Plugins
        plugins = info.get("Plugins", dict())
        
        # Key: bankInfo.Events
        events = info.get("Events", dict())
        
        # Key: bankInfo.DialogueEvents
        dialogues = info.get("DialogueEvents", dict())
        
        # Key: bankInfo.Busses
        busses = info.get("Busses", dict())
        
        # Key: bankInfo.AuxBusses
        auxes = info.get("AuxBusses", dict())
        
        # Key: bankInfo.GameParameters
        parameters = info.get("GameParameters", dict())
        
        # Key: bankInfo.Triggers
        triggers = info.get("Triggers", dict())
        
        # Key: bankInfo.StateGroups
        states = info.get("StateGroups", dict())
        
        # Key: bankInfo.SwitchGroups
        switches = info.get("SwitchGroups", dict())
        
        # Key: bankInfo.AcousticTextures
        textures = info.get("AcousticTextures", dict())
        
        # Key: pluginInfo
        libs = kwargs.get("pluginInfo", dict()).get("PluginLibs", list[dict]())
        libs = [PluginLibraryInfo(str(lib.get("LibName", "")), ShortID(lib.get("LibId", ShortID.get_null())),
                                  str(lib.get("Type", "")), str(lib.get("DLL", "")), str(lib.get("StaticLib", "")))
                for lib in libs]
        libs = tuple(libs)
        
        # Key: bankData
        bank_data = kwargs.get("bankData", dict())
        bank_data = SoundBankData(bank_data.get("data", ""), bank_data.get("size", 0))
        
        # Other data
        error_message = kwargs.get("error", "")
        platform = Name(kwargs.get("platform", dict()).get("name", Name.get_null()))
        language = Name(kwargs.get("language", Name.get_null()))
        
        event(SoundBankGenerationInfo(guid, name, path, file, short, bhash, btype, platform, language, media, externals,
                                      plugins, events, dialogues, busses, auxes, parameters, triggers, states, switches,
                                      textures, bank_data, libs, error_message))
    
    @callback
    def _on_generation_done(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `generationDone` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        items = [LogItem(severity=EnumStatics.from_value(ELogSeverity, item["severity"]), time=item["time"],
                         id=item["messageId"], description=item["message"]) for item in kwargs.get("logs", ())]
        event(tuple(items))
    
    def convert_external_sources(self, sources: ListOrTuple[ExternalSourceInfo]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_convertexternalsources.html \n
        Converts the external sources files for the project as detailed in the wsources file, and places
        them into either the default folder, or the folder specified by the output argument. External
        Sources are a special type of source that you can put in a Sound object in Wwise. It indicates
        that the real sound data will be provided at run time. While External Source conversion is also
        triggered by SoundBank generation, this operation can be used to process sources not contained in
        the Wwise Project. Please refer to Wwise SDK help page "Integrating External Sources".
        :param sources: An array of external sources files and corresponding arguments. Duplicates are ignored.
        :return: Whether the call succeeded.
        """
        args = {"sources": [source.dictionary for source in sources]}
        return self._client.call("ak.wwise.core.soundbank.convertExternalSources", args) is not None
    
    def generate(self, sound_banks: ListOrTuple[SoundBankInfo] = None, platforms: ListOrTuple[GUID | Name] = None,
                 languages: ListOrTuple[GUID | Name] = None, clear_audio_file_cache: bool = False,
                 write_to_disk: bool = False, rebuild_init_bank: bool = False) -> dict:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_generate.html \n
        Generate a list of SoundBanks with the import definition specified in the WAAPI request. If you
        do not write the SoundBanks to disk, subscribe to `ak.wwise.core.soundbank.generated` to receive
        SoundBank structure info and the bank data as base64. Note: This is a synchronous operation.
        :param sound_banks: A list of SoundBanks to generate.
        :param platforms: A list of platforms to generate. By default, all platforms will be generated.
        :param languages: A list of languages to generate. By default, all languages will be generated. To skip
                          languages, pass an empty set (`set()`) as the argument.
        :param clear_audio_file_cache: Deletes the content of the Wwise audio file cache folder prior to converting
                                       source files and generating SoundBanks, which ensures that all source files are
                                       reconverted.
        :param write_to_disk: Will write the sound bank and info file to disk.
        :param rebuild_init_bank: Use this option to force a rebuild of the Init bank for each specified platform.
        :return: The SoundBank generation log, as a dictionary. An empty dictionary means the log could not be
                  retrieved.
        """
        # Remove duplicates in collections
        sound_banks = list(dict.fromkeys(sound_banks)) if sound_banks is not None else list[SoundBankInfo]()
        platforms = list(dict.fromkeys(platforms)) if platforms is not None else list[GUID | Name]()
        languages = list(dict.fromkeys(languages)) if languages is not None else list[GUID | Name]()
        
        args = dict()
        
        if sound_banks:  # if valid and not empty
            args["soundbanks"] = [{"name": sound_bank} for sound_bank in sound_banks]
        else:
            args["rebuildSoundBanks"] = True
        
        if platforms:  # None = all platforms; no need to handle that here.
            args["platforms"] = platforms
        
        if languages:  # None = all languages; no need to handle that here.
            args["languages"] = languages
        else:  # no languages; user explicitly passed an empty set
            args["skipLanguages"] = True
        
        args["clearAudioFileCache"] = clear_audio_file_cache
        args["writeToDisk"] = write_to_disk
        args["rebuildInitBank"] = rebuild_init_bank
        
        log = self._client.call("ak.wwise.core.soundbank.generate", args).get("logs")
        return log if log is not None else dict()
    
    def get_inclusions(self, sound_bank: GUID | Name | ProjectPath) -> tuple[SoundBankInclusion, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_getinclusions.html \n
        Retrieves a SoundBank's inclusion list.
        :param sound_bank: The GUID, name, or project path of the SoundBank to get inclusions from.
        :return: An array of SoundBank inclusions.
        """
        if isinstance(sound_bank, Name):
            sound_bank = f"{EObjectType.SOUND_BANK.get_type_name()}:{sound_bank}"
        
        args = {"soundbank": sound_bank}
        results = self._client.call("ak.wwise.core.soundbank.getInclusions", args)
        
        if results is None:
            return tuple[SoundBankInclusion, ...]()
        
        results = results.get("inclusions", ())
        
        if not results:  # Empty.
            return tuple[SoundBankInclusion, ...]()
        
        inclusions = list[SoundBankInclusion]()
        
        for result in results:
            obj: GUID = GUID(result.get("object", GUID.get_null()))
            filters = tuple([EnumStatics.from_value(EInclusionFilter, _filter) for _filter in result.get("filter", ())])
            inclusions.append(SoundBankInclusion(obj, filters))
        
        return tuple[SoundBankInclusion, ...](inclusions)
    
    def process_definition_files(self, files: ListOrTuple[SystemPath]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_processdefinitionfiles.html \n
        Imports SoundBank definitions from the specified file. Multiple files can be specified. See the
        WAAPI log for status messages.
        :param files: An array of SoundBank definition files.
        :return: Whether the call succeeded.
        """
        args = {"files": [str(file) for file in list(dict.fromkeys(files))]}  # File paths should be unique.
        return self._client.call("ak.wwise.core.soundbank.processDefinitionFiles", args) is not None
    
    def set_inclusions(self, sound_bank: Name | GUID | ProjectPath, operation: EInclusionOperation,
                       inclusions: ListOrTuple[SoundBankInclusion]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_soundbank_setinclusions.html \n
        Modifies a SoundBank's inclusion list. The 'operation' argument determines how the 'inclusions'
        argument modifies the SoundBank's inclusion list; 'inclusions' may be added to / removed from /
        replace the SoundBank's inclusion list.
        :param sound_bank: The GUID, name, or project path of the SoundBank to add an inclusion to.
        :param operation: Determines how the 'inclusions' argument is used to modify the SoundBank's inclusion list.
        :param inclusions: An array of SoundBank inclusions.
        :return: Whether the call succeeded.
        """
        if isinstance(sound_bank, Name):
            sound_bank = f"{EObjectType.SOUND_BANK.get_type_name()}:{sound_bank}"
        args = {"soundbank": sound_bank, "operation": operation,
                "inclusions": [inclusion.dictionary for inclusion in list(dict.fromkeys(inclusions))]}
        return self._client.call("ak.wwise.core.soundbank.setInclusions", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EExternalAnalysisType, ELoudnessNormalizationType
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.conversion import Conversion


class ExternalSourceFile(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_externalsourcefile.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.EXTERNAL_SOURCE_FILE`.
    """
    analysis_type = WwiseProperty[EExternalAnalysisType]("AnalysisType", EExternalAnalysisType)
    conversion = WwiseProperty[Conversion]("Conversion", Conversion)
    loudness_normalization_target = WwiseProperty[float]("LoudnessNormalizationTarget", float)
    loudness_normalization_type = WwiseProperty[ELoudnessNormalizationType](
        "LoudnessNormalizationType", ELoudnessNormalizationType)
    override_conversion = WwiseProperty[bool]("OverrideConversion", bool)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour, EFadeMode, EFadeShape
from pywwise.objects.abc import WwiseObject


class MusicClip(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_musicclip.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MUSIC_CLIP`.
    """
    begin_trim_offset = WwiseProperty[float]("BeginTrimOffset", float)
    colour = WwiseProperty[EColour]("Color", EColour)
    end_trim_offset = WwiseProperty[float]("EndTrimOffset", float)
    fade_in_duration = WwiseProperty[float]("FadeInDuration", float)
    fade_in_mode = WwiseProperty[EFadeMode]("FadeInDuration", EFadeMode)
    fade_in_shape = WwiseProperty[EFadeShape]("FadeInShape", EFadeShape)
    fade_out_duration = WwiseProperty[float]("FadeOutDuration", float)
    fade_out_mode = WwiseProperty[EFadeMode]("FadeOutMode", EFadeMode)
    fade_out_shape = WwiseProperty[EFadeShape]("FadeOutShape", EFadeShape)
    highpass = WwiseProperty[int]("Highpass", int)
    lowpass = WwiseProperty[int]("Lowpass", int)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    play_at = WwiseProperty[float]("PlayAt", float)
    volume = WwiseProperty[float]("Volume", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.effect_slot import EffectSlot
from pywwise.objects.types.rtpc import Rtpc


class AudioDevice(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_audiodevice.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.AUDIO_DEVICE`.
    """
    bypass_effect = WwiseProperty[bool]("BypassEffect", bool)
    colour = WwiseProperty[EColour]("Color", EColour)
    effects = WwiseProperty[list[EffectSlot]]("Effects", list[EffectSlot])
    inclusion = WwiseProperty[bool]("Inclusion", bool)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    rtpc = WwiseProperty[list[Rtpc]]("RTPC", list[Rtpc])



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from typing import Union as _Union

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.blend_container import BlendContainer
from pywwise.objects.types.music_playlist_container import MusicPlaylistContainer
from pywwise.objects.types.music_segment import MusicSegment
from pywwise.objects.types.music_switch_container import MusicSwitchContainer
from pywwise.objects.types.random_sequence_container import RandomSequenceContainer
from pywwise.objects.types.sound import Sound
from pywwise.objects.types.state import State
from pywwise.objects.types.state_group import StateGroup
from pywwise.objects.types.switch import Switch
from pywwise.objects.types.switch_container import SwitchContainer
from pywwise.objects.types.switch_group import SwitchGroup

_AudioNodeType = _Union[
    RandomSequenceContainer, SwitchContainer, BlendContainer, Sound, MusicPlaylistContainer, MusicSegment,
    MusicSwitchContainer]

_AudioNodeTypeTuple = (
    RandomSequenceContainer, SwitchContainer, BlendContainer, Sound, MusicPlaylistContainer, MusicSegment,
    MusicSwitchContainer)


class MultiSwitchEntry(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_multiswitchentry.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MULTI_SWITCH_ENTRY`.
    """
    audio_node = WwiseProperty[_AudioNodeType]("AudioNode", _AudioNodeTypeTuple)
    entry_path = WwiseProperty[tuple[SwitchGroup | Switch | StateGroup | State, ...]]("EntryPath", tuple)
    probability = WwiseProperty[int]("Probability", int)
    weight = WwiseProperty[int]("Weight", int)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class Modifier(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_modifier.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MODIFIER`.
    """
    enabled = WwiseProperty[bool]("Enabled", bool)
    max = WwiseProperty[float]("Max", float)
    min = WwiseProperty[float]("Min", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.decorators import callback
from pywwise.enums import ECaptureLogItemType, ECaptureLogSeverity
from pywwise.primitives import GameObjectID, GUID, Name, PlayingID, ShortID
from pywwise.statics import EnumStatics
from pywwise.structs import CaptureLogItem


class CaptureLog:
    """ak.wwise.core.profiler.capture_log"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.item_added = _RefEvent(CaptureLogItem)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_capturelog_itemadded.html
        \nSent when a new entry is added to the capture log.
        \n**Event Data**:
        \n- An instance of CaptureLogItem, which contains information such as type, time, severity, etc.
        """
        
        self._item_added = self._client.subscribe("ak.wwise.core.profiler.captureLog.itemAdded", self._on_item_added)
    
    @callback
    def _on_item_added(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `itemAdded` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        item_type = EnumStatics.from_value(ECaptureLogItemType, str(kwargs["type"]).replace(" ", ""))
        time_seconds = kwargs["time"]
        description = kwargs["description"]
        severity = EnumStatics.from_value(ECaptureLogSeverity, kwargs["severity"])
        wwise_obj_id = GUID(kwargs["objectId"]) if "objectId" in kwargs else GUID.get_null()
        wwise_obj_name = Name(kwargs["objectName"]) if "objectName" in kwargs else Name.get_null()
        wwise_obj_short = ShortID(kwargs["objectShortId"]) if "objectShortId" in kwargs else ShortID.get_null()
        game_obj_id = GameObjectID(kwargs["gameObjectId"]) if "gameObjectId" in kwargs else GameObjectID.get_null()
        game_obj_name = Name(kwargs["gameObjectName"]) if "gameObjectName" in kwargs else Name.get_null()
        playing_id = PlayingID(kwargs["playingId"]) if "playingId" in kwargs else PlayingID.get_null()
        error_code_name = kwargs.get("errorCodeName", "")
        event(CaptureLogItem(item_type, time_seconds, description, severity, wwise_obj_id, wwise_obj_name,
                             wwise_obj_short, game_obj_id, game_obj_name, playing_id, error_code_name))



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class Path2D(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_path2d.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.PATH_2D`.
    """
    append_offset = WwiseProperty[int]("AppendOffset", int)
    duration = WwiseProperty[int]("Duration", int)
    flags = WwiseProperty[int]("Flags", int)
    linear_time = WwiseProperty[bool]("LinearTime", bool)
    random_x = WwiseProperty[float]("RandomX", float)
    random_y = WwiseProperty[float]("RandomY", float)
    random_z = WwiseProperty[float]("RandomZ", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from enum import Enum as _Enum
from typing import Any as _Any, Generic as _Generic, Self as _Self, Type as _Type, TypeVar as _TypeVar

from pywwise.aliases import SystemPath
from pywwise.modules import LazyModule
from pywwise.primitives import GUID
from pywwise.statics import EnumStatics
from pywwise.structs import WwiseObjectInfo

_pywwise_objects = LazyModule("pywwise.objects")

_T = _TypeVar("_T")


class WwiseProperty(_Generic[_T]):
    """A descriptor for Wwise properties, references, and lists. Supports the `WwiseProperty[Type]` syntax to allow
    auto-completion and suggestions in interactive environments."""
    
    def __init__(self, name: str, etype: _Type[_T] | tuple[_Type[_T]]):
        """
        Initializer.
        :param name: The name of the property.
        :param etype: The type of the property. If using the `WwiseProperty[Type]` syntax, this parameter MUST match
        the type specified in brackets. Due to a limitation in the Python syntax, `etype` cannot be used for type hints.
        """
        self._name = name
        self._type = etype
    
    def __get__(self, instance, owner) -> _T | str:
        """
        Getter.
        :param instance: The caller.
        :param owner: The owner class.
        :return: The current value.
        """
        if instance is None:
            return self._name
        
        getter = getattr(instance, "get_property")
        
        if getter is None:
            raise TypeError("Encapsulators of `WwiseProperty` must implement the `get_property` function.")
        
        ak = getattr(instance, "_ak")
        
        if ak is None:
            raise TypeError("Encapsulators of `WwiseProperty` must define a protected `WwiseConnection` named `_ak`.")
        
        value = getter(self._name)
        
        _type = self._type if self._type is not _Self else instance.__class__  # Class is referencing itself.
        
        match _type:  # Decide on what kind of object to return.
            
            case _ if isinstance(_type, tuple) and isinstance(value, dict):
                query = f"$ from object \"{value.get('id', GUID.get_null())}\" take 1"
                info_tuple: tuple[WwiseObjectInfo, ...] = ak.wwise.core.object.get(query)
                if info_tuple is None or not info_tuple:  # Invalid or empty.
                    raise ValueError(f"Invalid object returned for property `{self._name}`. Either `None` or empty.")
                info_tuple: WwiseObjectInfo = info_tuple[0]  # Still tuple; convert to single value (GUIDs are unique).
                return info_tuple.type.get_class()(info_tuple.guid, ak)
            
            case _ if (_type is list or _type is tuple) and (isinstance(value, list) or isinstance(value, tuple)):
                info_list = list[WwiseObjectInfo]()  # Get `WwiseObjectInfo` objs.
                for dictionary in value:
                    query = f"$ from object \"{dictionary.get('id', GUID.get_null())}\" take 1"
                    info_tuple: tuple[WwiseObjectInfo, ...] = ak.wwise.core.object.get(query)
                    if info_tuple is not None and info_tuple:  # Valid, not empty.
                        info_list.append(info_tuple[0])  # GUIDs are unique, so there is only one value in that tuple.
                return tuple(info.type.get_class()(info.guid, ak) for info in info_list)
            
            case _ if issubclass(_type, _pywwise_objects.WwiseObject) and isinstance(value, dict):  # WwiseObject
                return _type(value.get("id", GUID.get_null()), ak)
            
            case _ if issubclass(_type, _Enum):  # Any generic enum.Enum, but usually a pywwise.enums type.
                return EnumStatics.from_value(_type, value)
            
            case _:  # Anything else, including GUID, ProjectPath, GameObjectID, SystemPath, etc.
                return _type(value)  # PyCharm might throw a warning here about a missing `ak`; false negative.
    
    def __set__(self, instance: _Any, value: _T):
        """
        Setter.
        :param instance: The caller.
        :param value: The new value.
        """
        setter = getattr(instance, "set_property")
        
        if setter is None:
            raise TypeError("Encapsulators of WwiseProperty must implement the `set_property` function.")
        
        if self._type == SystemPath:  # pathlib.Path (SystemPath) is not JSON-serializable; convert to `str`.
            value = str(value)
        elif issubclass(value.__class__, _pywwise_objects.WwiseObject) or isinstance(value, WwiseObjectInfo):
            value = value.guid
        
        instance.set_property(self._name, value, isinstance(value, GUID) or value is None)
    
    @property
    def type(self) -> _Type[_T]:
        """:return: The type of this property."""
        return self._type



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour, EControlSurfaceBindingGroupType, EControlSurfaceBindingTargetType
from pywwise.objects.abc import WwiseObject


class ControlSurfaceBindingGroup(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_controlsurfacebindinggroup.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.CONTROL_SURFACE_BINDING_GROUP`.
    """
    colour = WwiseProperty[EColour]("Color", EColour)
    group_type = WwiseProperty[EControlSurfaceBindingGroupType]("GroupType", EControlSurfaceBindingGroupType)
    hardware_controller_key = WwiseProperty[str]("HardwareControllerKey", str)
    object_index_in_view = WwiseProperty[int]("ObjectIndexInView", int)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    target_class_id = WwiseProperty[int]("TargetClassId", int)
    target_name = WwiseProperty[str]("TargetName", str)
    target_type = WwiseProperty[EControlSurfaceBindingTargetType]("TargetType", EControlSurfaceBindingTargetType)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.enums import EGameParameterValueChangeAction
from pywwise.primitives import GUID, Name, ProjectPath


class GameParameter:
    """ak.wwise.core.GameParameter"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
    
    def set_range(self, game_parameter: GUID | Name | ProjectPath, min_value: float, max_value: float,
                  on_curve_update: EGameParameterValueChangeAction) -> bool:
        """
        https://www.audiokinetic.com/en/library/edge/?source=SDK&id=ak_wwise_core_gameparameter_setrange.html \n
        Sets the Min and Max properties on a Game Parameter. Modifies the RTPC curves and blend tracks that use this
        Game Parameter for their X axis.
        :param game_parameter: The GUID, Name, or the project path of the Game Parameter.
        :param min_value: The minimum value of the Game Parameter.
        :param max_value: The maximum value of the Game Parameter.
        :param on_curve_update: Modifying the Min or Max value of a Game Parameter affects the RTPC curves and blend
                                tracks that use that Game Parameter for their X axis. Two actions are possible, Stretch
                                or PreserveX.
        :return: True if the call was successful, False otherwise.
        """
        args = {"object": game_parameter, "min": min_value, "max": max_value,
                "onCurveUpdate": on_curve_update}
        
        return self._client.call("ak.wwise.core.gameParameter.setRange", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from typing import Union as _Union

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.curve import Curve
from pywwise.objects.types.game_parameter import GameParameter
from pywwise.objects.types.midi_parameter import MidiParameter
from pywwise.objects.types.modulator_envelope import ModulatorEnvelope
from pywwise.objects.types.modulator_lfo import ModulatorLfo
from pywwise.objects.types.modulator_time import ModulatorTime

_ControlInputType = _Union[ModulatorLfo, ModulatorEnvelope, ModulatorTime, GameParameter, MidiParameter]

_ControlInputTypeTuple = (ModulatorLfo, ModulatorEnvelope, ModulatorTime, GameParameter, MidiParameter)


class Rtpc(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_rtpc.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.RTPC`.
    """
    colour = WwiseProperty[EColour]("Color", EColour)
    control_input = WwiseProperty[_ControlInputType]("ControlInput", _ControlInputTypeTuple)
    curve = WwiseProperty[Curve]("Curve", Curve)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    property_name = WwiseProperty[str]("PropertyName", str)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class ObjectSettingAssoc(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_objectsettingassoc.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.OBJECT_SETTING_ASSOC`.
    """
    continue_play = WwiseProperty[bool]("ContinuePlay", bool)
    fade_in_time = WwiseProperty[float]("FadeInTime", float)
    fade_out_time = WwiseProperty[float]("FadeOutTime", float)
    first_occurrence_only = WwiseProperty[bool]("FirstOccurenceOnly", bool)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EMotionChannelConfig, EMotionDriver
from pywwise.objects.abc import WwiseObject


class MotionSource(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_motion_source.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`.
    """
    channel_1 = WwiseProperty[EMotionDriver]("Channel1", EMotionDriver)
    channel_2 = WwiseProperty[EMotionDriver]("Channel2", EMotionDriver)
    channel_3 = WwiseProperty[EMotionDriver]("Channel3", EMotionDriver)
    channel_4 = WwiseProperty[EMotionDriver]("Channel4", EMotionDriver)
    channel_config = WwiseProperty[EMotionChannelConfig]("ChannelConfig", EMotionChannelConfig)
    driver_a = WwiseProperty[float]("DriverA", float)
    driver_b = WwiseProperty[float]("DriverB", float)
    driver_c = WwiseProperty[float]("DriverC", float)
    driver_d = WwiseProperty[float]("DriverD", float)
    driver_e = WwiseProperty[float]("DriverE", float)
    driver_f = WwiseProperty[float]("DriverF", float)
    driver_g = WwiseProperty[float]("DriverG", float)
    driver_h = WwiseProperty[float]("DriverH", float)
    high_1 = WwiseProperty[EMotionDriver]("High1", EMotionDriver)
    high_2 = WwiseProperty[EMotionDriver]("High2", EMotionDriver)
    left_trigger = WwiseProperty[EMotionDriver]("LeftTrigger", EMotionDriver)
    low_1 = WwiseProperty[EMotionDriver]("Low1", EMotionDriver)
    low_2 = WwiseProperty[EMotionDriver]("Low2", EMotionDriver)
    right_trigger = WwiseProperty[EMotionDriver]("RightTrigger", EMotionDriver)
